@@ -1901,12 +1901,19 @@ class GitHandler(object):
                 username, password = pwmgr.find_user_password(
                     self._http_auth_realm, str_uri,
                 )
+                # NB: probably bytes here
             elif auth is not None:
                 username, password = auth
-                username = username.decode('utf-8')
-                password = password.decode('utf-8')
+                # NB: probably string here
             else:
                 username, password = pwmgr.find_stored_password(str_uri)
+                # NB: probably string here
+
+            if isinstance(username, bytes):
+                username = username.decode('utf-8')
+
+            if isinstance(password, bytes):
+                password = password.decode('utf-8')
 
             return (
                 client.HttpGitClient(
